"""
Contains the configuration for the ixoncdkingress webserver
"""
import enum
import json
import sys

from functools import reduce
from logging import Formatter, Logger, StreamHandler, getLogger
from typing import Mapping, Optional, Union, Dict, Any, Tuple

class InvalidConfigException(Exception):
    """
    Exception for invalid values in the config
    """

class WsgiProvider(enum.Enum):
    """
    Provider to use to turn a WSGI application into a (HTTP) server
    """
    WSGIREF = 'wsgiref'
    BJOERN = 'bjoern'
    WAITRESS = 'waitress'
    GUNICORN = 'gunicorn'
    CHERRYPY = 'cherrypy'
    MEINHELD = 'meinheld'

class Config:
    """
    Configuration for the webserver
    """
    @classmethod
    def from_environ(cls, environ: Mapping[str, str]) -> 'Config':
        """
        Parses an environment disctionary and generates a config based on it
        """

        res = cls()

        res.production_mode = load_bool_from_environment(environ, 'PRODUCTION_MODE', False)

        res.http_server_bind = environ.get('HTTP_SERVER_BIND', '127.0.0.1')
        res.http_server_port = int(environ.get('HTTP_SERVER_PORT', '8020'))

        res.wsgi_provider = WsgiProvider(environ.get('WSGI_PROVIDER', 'wsgiref'))

        res.logger_log_level = environ.get('LOGGER_LOG_LEVEL', 'INFO')
        res.logger_format = environ.get(
            'LOGGER_FORMAT', '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )

        res.version = load_str_file(environ.get('VERSION_PATH'), 'dev')

        res.cbc_path = environ.get('CBC_PATH', '/user_scripts')

        res.api_client_base_url = environ.get('API_CLIENT_BASE_URL', 'https://api.ayayot.com/')

        return res

    production_mode: bool

    http_server_bind: str
    http_server_port: int
    wsgi_provider: WsgiProvider

    logger_log_level: Union[int, str]
    logger_format: str

    version: str
    cbc_path: str
    api_client_base_url: str

    _logger: Optional[Logger]

    def __init__(self) -> None:
        self._logger = None

    def get_logger(self) -> Logger:
        """
        Returns a logger
        """
        if self._logger is None:
            logger = getLogger('ixoncdkingress')
            logger.setLevel(self.logger_log_level)

            handler = StreamHandler(sys.stdout)
            handler.setFormatter(Formatter(self.logger_format))
            logger.addHandler(handler)

            self._logger = logger

        return self._logger

def load_str_file(
        file_path: Optional[str], default: str
    ) -> str:
    """
    Loads a string from a file, returns default if no path is given
    """
    if not file_path:
        return default

    try:
        with open(file_path, mode='r', encoding='utf-8') as file:
            string = file.read().rstrip()
            if not string:
                return default
        return string
    except IOError:
        return default

def to_str_value(acc: Dict[str, str], val: Tuple[str, Any]) -> Dict[str, str]:
    """
    Reducer function to convert a dict entry value to string.
    """
    return acc | { val[0]: str(val[1])}

def load_bool_from_environment(environ: Mapping[str, str], key: str, default: bool) -> bool:
    """
    Loads a boolean value from the environment
    """
    val_str = environ.get(key, default)
    if val_str is default or val_str == '':
        return default

    if val_str not in ('False', 'True'):
        raise InvalidConfigException(f'Invalid boolean value: {val_str}')

    return 'True' == val_str

def load_json_file(
        file_path: Optional[str], default: Dict[str, str]
    ) -> Dict[str, str]:
    """
    Loads a Dict from a JSON file, converting all numeric values to string. Returns default if no
    path is given, file_path does not exist, or file_path does not contain a JSON file.
    """
    if not file_path:
        return default

    try:
        with open(file_path, mode='r', encoding='utf-8') as file:
            result = json.load(file, parse_int=str, parse_float=str)
            if not result:
                return default

        values = result['values']

        return reduce(to_str_value, values.items(), {})
    except IOError:
        return default



"""
Context.
"""
from typing import Any, Dict, Optional, Set

from ixoncdkingress.cbc.api_client import ApiClient
from ixoncdkingress.cbc.document_db_client import DocumentDBClient

class CbcResource:
    """
    Describes an IXAPI resource.
    """
    public_id: str
    name: str
    custom_properties: Dict[str, Any]
    permissions: Optional[Set[str]]

    def __init__(
            self,
            public_id: str,
            name: str,
            custom_properties: Dict[str, Any],
            permissions: Optional[Set[str]]
        ) -> None:
        self.public_id = public_id
        self.name = name
        self.custom_properties = custom_properties
        self.permissions = permissions

    def __repr__(self) -> str:
        return (
            '<CbcResource'
            f' public_id={self.public_id},'
            f' name={self.name},'
            f' custom_properties={repr(self.custom_properties)},'
            f' permissions={repr(self.permissions)},'
            f'>'
        )

class CbcContext:
    """
    The context for a backend component.
    """
    config: Dict[str, str]
    api_client: ApiClient
    document_db_client: Optional[DocumentDBClient] = None
    user: Optional[CbcResource] = None
    company: Optional[CbcResource] = None
    asset: Optional[CbcResource] = None
    agent: Optional[CbcResource] = None

    @property
    def agent_or_asset(self) -> CbcResource:
        """
        Return either an Agent or an Asset resource, depending on what's available. If both are
        available in the context, returns the Asset resource.
        """
        if self.asset:
            return self.asset

        assert self.agent
        return self.agent

    def __init__(
            self,
            config: Dict[str, str],
            api_client: ApiClient,
            document_db_client: Optional[DocumentDBClient] = None,
            user: Optional[CbcResource] = None,
            company: Optional[CbcResource] = None,
            asset: Optional[CbcResource] = None,
            agent: Optional[CbcResource] = None,
            **kwargs: Any
        ) -> None:
        del kwargs

        self.config = config
        self.api_client = api_client
        self.document_db_client = document_db_client
        self.user = user
        self.company = company
        self.asset = asset
        self.agent = agent

    def __repr__(self) -> str:
        return (
            f'<CbcContext'
            f' config={repr(self.config)},'
            f' api_client={repr(self.api_client)},'
            f' document_db_client={repr(self.document_db_client)},'
            f' user={repr(self.user)},'
            f' company={repr(self.company)},'
            f' asset={repr(self.asset)},'
            f' agent={repr(self.agent)},'
            f'>'
        )



import mock
from ixtest.mock import Mut, create_mock

from ixoncdkingress import utils as sut

def test_handle_reload_exception(monkeypatch):
    mut = Mut(sut.handle_exception)

    mut.exception = create_mock(name='exception')
    mut.exception.__traceback__ = 'my tracback'

    mut.create_test_suite(monkeypatch)

    mut()

    assert [
        mock.call(None, mut.exception, mut.exception.__traceback__)
    ] == sut.traceback.print_exception.call_args_list
    assert sut.ResponseCode.INTERNAL_ERROR == mut.response.status_code



"""
Module containing a client that is a abstraction of the MongoClient.
"""

from dataclasses import dataclass
from typing import Any, Dict, Iterable, Mapping, Optional

from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.results import (
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)

TIMEOUT: int = 30000

DocumentType = Dict[str, Any]


@dataclass(frozen=True)
class DocumentDBAuthentication:
    """
    Dataclass containing authentication values for the Document DB
    """

    username: str
    password: str


@dataclass(frozen=True)
class DocumentDBConnection:
    """
    Dataclass containing host values for the Document DB
    """

    host: str
    port: int


class DocumentDBClient:
    """
    Class for a document db client, this client connects with a document database.
    """

    _mongo_client: MongoClient[DocumentType]
    _database: str
    _collection: str
    _authentication: DocumentDBAuthentication

    def __init__(
        self,
        connection: DocumentDBConnection,
        database: str,
        collection: str,
        authentication: DocumentDBAuthentication,
    ) -> None:
        """
        database: Company PublicId
        collection: CBC PublicId
        """

        self._database = database
        self._collection = collection
        self._authentication = authentication

        self._mongo_client = MongoClient(
            host=connection.host,
            port=connection.port,
            username=self._authentication.username,
            password=self._authentication.password,
            timeoutMS=TIMEOUT,
            tls=True,
        )

    def insert_one(self, document: DocumentType) -> InsertOneResult:
        """
        Insert a single document.
        """
        return self._get_collection().insert_one(document)

    def insert_many(self, documents: Iterable[DocumentType]) -> InsertManyResult:
        """
        Insert an iterable of documents.
        """
        return self._get_collection().insert_many(documents)

    def update_one(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update a single document matching the filter.
        """
        return self._get_collection().update_one(filter_map, update)

    def update_many(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update one or more documents that match the filter.
        """
        return self._get_collection().update_many(filter_map, update)

    def delete_one(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete a single document matching the filter.
        """
        return self._get_collection().delete_one(filter_map)

    def delete_many(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete one or more documents matching the filter.
        """
        return self._get_collection().delete_many(filter_map)

    def find_one(
        self, *args: Any, filter_map: Optional[Any], **kwargs: Any
    ) -> Optional[DocumentType]:
        """
        Get a single document from the database.
        """
        return self._get_collection().find_one(filter_map, args, kwargs)

    def find(self, *args: Any, **kwargs: Any) -> Cursor[DocumentType]:
        """
        Query the database.
        """
        return self._get_collection().find(args, kwargs)

    def _get_collection(self, collection_name: Optional[str] = None) -> Collection[DocumentType]:
        return self._mongo_client.get_database(self._database).get_collection(
            collection_name or self._collection
        )

    def __repr__(self) -> str:
        return (
            '<DocumentDBClient'
            f' database={self._database},'
            f' collection={self._collection},'
            '>'
        )



"""
The __main__ module lets the user start the ingress from the command line
"""
from typing import List

import os, sys

from ixoncdkingress.webserver.servlet import Servlet
from ixoncdkingress.webserver.config import Config
from ixoncdkingress.webserver import wsgi

def main(argv: List[str]) -> int:
    """
    Parses the arguments and runs the provider
    """
    del argv # Configuration goes via environment

    config = Config.from_environ(os.environ)
    servlet = Servlet(config)

    try:
        wsgi.run_server(config, servlet)
        return 0
    except KeyboardInterrupt:
        ...

    return 0

if '__main__' == __name__: # pragma: no cover
    sys.exit(main(sys.argv[1:]))
